package org.implab.gradle.mercurial;

import java.io.Serializable;
import java.util.Objects;
import java.util.Optional;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import javax.annotation.Nonnull;

public class SemVersion implements Serializable, Comparable<SemVersion> {

    private static final long serialVersionUID = 1L;

    private static Pattern semVerPattern = Pattern.compile(
            "^(0|[1-9]\\d*)\\.(0|[1-9]\\d*)\\.(0|[1-9]\\d*)(?:-((?:0|[1-9]\\d*|\\d*[a-zA-Z-][0-9a-zA-Z-]*)(?:\\.(?:0|[1-9]\\d*|\\d*[a-zA-Z-][0-9a-zA-Z-]*))*))?(?:\\+([0-9a-zA-Z-]+(?:\\.[0-9a-zA-Z-]+)*))?$");

    private static Pattern semVerLaxPattern = Pattern.compile(
            "^(0|[1-9]\\d*)(?:\\.(0|[1-9]\\d*)(?:\\.(0|[1-9]\\d*))?)?(?:-((?:0|[1-9]\\d*|\\d*[a-zA-Z-][0-9a-zA-Z-]*)(?:\\.(?:0|[1-9]\\d*|\\d*[a-zA-Z-][0-9a-zA-Z-]*))*))?(?:\\+([0-9a-zA-Z-]+(?:\\.[0-9a-zA-Z-]+)*))?$");

    private int major;

    private int minor;

    private int patch;

    private String suffix;

    private String meta;

    public SemVersion() {
    }

    public SemVersion(int major, int minor, int patch, String suffix, String meta) {
        this.major = major;
        this.minor = minor;
        this.patch = patch;
        this.suffix = suffix;
        this.meta = meta;
    }

    public SemVersion(int major, int minor, int patch) {
        this(major, minor, patch, null, null);
    }

    public SemVersion(SemVersion src) {
        this(src.major, src.minor, src.patch, src.suffix, src.meta);
    }

    public int getMajor() {
        return major;
    }

    public void setMajor(int major) {
        this.major = major;
    }

    public int getMinor() {
        return minor;
    }

    public void setMinor(int minor) {
        this.minor = minor;
    }

    public int getPatch() {
        return patch;
    }

    public void setPatch(int patch) {
        this.patch = patch;
    }

    public Optional<String> getSuffix() {
        return Optional.ofNullable(suffix);
    }

    public void setSuffix(String suffix) {
        this.suffix = suffix;
    }

    public Optional<String> getMeta() {
        return Optional.ofNullable(meta);
    }

    public void setMeta(String meta) {
        this.meta = meta;
    }

    public boolean isRelease() {
        return getSuffix().isEmpty();
    }

    @Override
    public String toString() {
        StringBuilder sb = new StringBuilder();

        sb.append(major).append(".").append(minor).append(".").append(patch);
        getSuffix().ifPresent(s -> sb.append("-").append(s));
        getMeta().ifPresent(m -> sb.append("+").append(m));

        return sb.toString();
    }

    @Override
    public int compareTo(SemVersion other) {
        int res = Integer.compare(major, other.major);
        if (res != 0)
            return res;
        res = Integer.compare(minor, other.minor);
        if (res != 0)
            return res;
        res = Integer.compare(patch, other.patch);

        return getSuffix()
                // if we have suffix, chech another suffix, if there is no another suffix
                // then another version is bigger: 1.0.0-rc < 1.0.0
                .map(s1 -> other.getSuffix().map(s2 -> s1.compareTo(s2)).orElse(-1))
                // if we don't have the suffix
                .orElse(other.getSuffix().isPresent() ? 1 : 0);

    }

    public SemVersion release() {
        return new SemVersion(major, minor, patch, null, null);
    }

    public SemVersion suffix(String suffix) {
        return new SemVersion(major, minor, patch, suffix, meta);
    }

    public SemVersion suffix() {
        return new SemVersion(major, minor, patch, null, meta);
    }

    public SemVersion meta(String meta) {
        return new SemVersion(major, minor, patch, suffix, meta);
    }

    public SemVersion meta() {
        return new SemVersion(major, minor, patch, suffix, null);
    }

    public SemVersion patch(int patch) {
        return new SemVersion(major, minor, patch, suffix, meta);
    }

    public SemVersion minor(int minor) {
        return new SemVersion(major, minor, patch, suffix, meta);
    }

    public SemVersion major(int major) {
        return new SemVersion(major, minor, patch, suffix, meta);
    }

    public SemVersion addPatch(int number) {
        return new SemVersion(major, minor, patch + number, suffix, meta);
    }

    public SemVersion addMinor(int number) {
        return new SemVersion(major, minor + number, patch, suffix, meta);
    }

    public SemVersion addMajor(int number) {
        return new SemVersion(major + number, minor, patch, suffix, meta);
    }

    public SemVersion nextMinor() {
        return new SemVersion(major, minor + 1, 0, null, meta);
    }

    public SemVersion nextMajor() {
        return new SemVersion(major + 1, 0, 0, null, meta);
    }

    public static Optional<SemVersion> tryParse(@Nonnull String str) {
        Objects.requireNonNull(str);

        Matcher match = semVerPattern.matcher(str);

        return match.matches() ? Optional.of(new SemVersion(Integer.parseInt(match.group(1)),
                Integer.parseInt(match.group(2)), Integer.parseInt(match.group(3)), match.group(4), match.group(5)))
                : Optional.empty();
    }

    public static Optional<SemVersion> tryParseLax(@Nonnull String str) {
        Objects.requireNonNull(str);

        Matcher match = semVerLaxPattern.matcher(str);

        return match.matches() ? Optional.of(new SemVersion(Integer.parseInt(match.group(1)),
                Integer.parseInt(Optional.ofNullable(match.group(2)).orElse("0")),
                Integer.parseInt(Optional.ofNullable(match.group(3)).orElse("0")), match.group(4), match.group(5)))
                : Optional.empty();
    }

    public static SemVersion toSemVersion(Object value, boolean strict) {
        if (value == null) {
            return new SemVersion();
        } else if (value instanceof SemVersion) {
            return new SemVersion((SemVersion) value);
        } else {
            Optional<SemVersion> semver = strict ? tryParse(value.toString()) : tryParseLax(value.toString());
            return semver.orElseThrow(() -> new IllegalArgumentException(String.format("Failed to parse semver '%s', strict=%s", value, strict)));
        }
    }

    public static SemVersion toSemVersion(Object value) {
        return toSemVersion(value, false);
    }
}
